
from qckt.QCkt import QCkt
from qckt.Gates import define_gate, get_gates_list
from qckt.Job import Job
from qckt.qException import QCktException

__version__ = '3.1'
__author__ = 'Atul Varshneya'
__author_email__ = 'atul.varshneya@gmail.com'
__email__ = __author_email__
__url__ = 'https://github.com/atulvarshneya/quantum-computing'
